*- Name.....Combobox with Auto Fill Search (ComboAFS)
*- Author...Todd Kreuter
*- For......dBASE Plus, dB2K

*- Description
   // A combobox providing an incremental search to items 
   // available in the dropDownList. Combobox can act
   // as either a dropDown or dropDownList style combobox.
   // The dropDownList style is simulated by restricting
   // user input to existing items via the key event.

	// This custom class works around a few problems found in
   // the stock combobox class:
   	// Datalinked value not changed when control value is
      // changed via arrow keys.

      // Datalinked value remaining blank when blanking the
      // control value.

      // Edit value not selected when control has focus and
      // rowset is navigated, saved, or abandoned.

	// Includes Refresh custom method which is fired from the
   // onTimer event and is primarily a means for refreshing
   // the dataSource as needed. To use refresh, a function 
   // pointer must exist in the form class and must be in the
   // following format (added using source editor).
   	//
      // Function <Control Name>_<Method Name>
      //		do somthing
      //		return

      // In the following example of using refresh, the refreshKey
      // property is used to determine when the initDataSource
      // needs to be called. The dataSource rowset is constrained
      // with the indexName and rangekey.
      //
      // Function ComboAFS1_refresh
      //	   if this.refreshKey # form.rowset.fields['Type'].value
      //       this.refeshKey = form.rowset.fields['Type'].value

      //       this.indexName = "Type"
      //       this.rangeKey  = this.refreshKey
      //
      //       this.initDataSource()
      //    endif
      //    return

	// When the dataSource is derived from a field value,
   // the list items are not trimmed, making the auto-fill
   // search feature less desireable. Refresh the dropDownList
   // using the initDataSource method to get a trimmed list.

	// The custom property, refreshDDL, determines whether or not
   // the dropDownList is refreshed using initDataSource before
   // the control first receieves focus (evaluated in when event).
	// This provides automatic refresh of the dropDownList. Setting
   // refreshDDL to false (onOpen) will prevent this refresh.

	// When overriding control events, be sure to make a call
   // back to the ComboAFS event. Otherwise, the control will
   // not work corrrectly.

*- External Use Properties
// styleDDL			-	Whether to simulate dropDownList style,
	//						restricts user input to items in the list.
   //
// allowBlank		-  Whether to allow user to blank the value
	//						when styleDDL
   //
// autoDropOnKey	-	Whether to drop the list when a key is
	//						pressed (once per focus)
   //
// refreshDDL		-  Forces the dropDownList to be refreshed once
	//						using the initDataSource method (in when).
   //
//	refreshKey		-	Use with refresh event code (optional).
	//
//	indexName		-	Used in InitDataSource, the indexName to
	//						apply to the dataSource rowset.
	//
// rangeKey			-	Used in InitDataSource, the rangeKey to 
	//						apply to the dataSource rowset.setRange().
   //
//	filter			-	Used in InitDataSource, the filter to apply
	//						to the dataSource rowset.

*- External Use Methods
// InitDataSource -	Rebuilds the dataSource from a field object
	//						applying any indexName, rangeKey, or filter
   //						settings, and trimming the values.
	//
// InsertString	-	Inserts a string into the dropDown list at
	//						a specified index location (zero based)

// Refresh			-	Provides a means of refreshing the controls
	//						dataSource.

*- History
// 2002.0625 - Initial
// 2002.0628 - BugFix: Combobox value not recognized as empty
	//				when control is empty (reverts to prior value).
// 2002.0628 - Added allowBlank property to allow user to
	//				blank value when styleDDL
// 2002.0827 - Added InitDataSource method to rebuild dataSource
	//				derived from field value (trimmed).
   //				Added indexName, rangeKey, and filter properties
   //				which are used by initDataSource to constrain and
   //				order the dataSource.
// 2002.1228 - Removed all API PeekMessage related code.
	//				Rewrote key event code, added onKey code.
   //				Added Refresh method and refreshKey
   //				Added bookMark, state, and modified properties to
   //				fix problem where changing value via arrow keys was
   //				not changing the control.
// 2002.1230 - Added refreshDDL to force refresh of dropDownList.
// 2003.0126 - Changes to initDataSource method to work with
	// 			temporary tables and lookupSql. Option to requery
   //				or refresh dataSource rowset to ensure list is
   //				is updated from existing table rows.
// 2003.1216 - Renamed function bugfix to workaround.
	//				Changes to workaround method.

*----------------

#define LOWORD(x)           			(MOD(x,2^16))
#define HIWORD(x)           			(BITRSHIFT(MOD(x,2^32),16))

#define EM_SETMARGINS         		0x00D3	// Yes
#define EC_LEFTMARGIN       			0x0001	// Yes
#define EM_GETSEL               		0x00B0	// Yes
#define EM_SETSEL             		0x00B1	// Yes

#define CB_ADDSTRING          		0x0143	// Yes
#define CB_INSERTSTRING             0x014A	// Yes
#define CB_DELETESTRING             0x0144
#define CB_RESETCONTENT             0x014B	// Yes
#define CB_GETITEMHEIGHT            0x0154
#define CB_GETCOUNT                 0x0146
#define CB_GETCURSEL                0x0147
#define CB_SETCURSEL                0x014E	// Yes
#define CB_FINDSTRING               0x014C	// Yes
#define CB_FINDSTRINGEXACT          0x0158
#define CB_SELECTSTRING             0x014D
#define CB_GETLBTEXT                0x0148
#define CB_GETDROPPEDSTATE          0x0157	// Yes
#define CB_SHOWDROPDOWN             0x014F	// Yes

class ComboAFS(parentObj) of Combobox(parentObj)
	protect style

	*- API for edit and combbox messages
   if TYPE("SendMessageN") # "FP"
      extern   CLONG SendMessageN(CHANDLE, CUINT, CUINT, CLONG) User32 ;
               from "SendMessageA"
   endif

   if TYPE("SendMessageC") # "FP"
      extern	CLONG SendMessageC(CHANDLE, CUINT, CUINT, CSTRING) User32 ;
      			from "SendMessageA"
   endif

	*- API to get the hwnd of the edit control used in combbox
   if TYPE('FindWindowEx') # 'FP'
      extern   CHANDLE FindWindowEx(CHANDLE, CHANDLE, ;
               CSTRING, CSTRING) User32 from 'FindWindowExA'
   endif

	*- Stock properties
	with (this)
      height			= 20
      width				= 100
      fontSize			= 8
      metric			= 6	// Pixels
      style				= 1	// DropDown
      dropDownHeight = 132
   endwith

 	*- Internal Use properties
   this.hwndEdit		= 0			// Windows Handle of edit control
   this.autoDropFlag	= false		// Flag for auto dropping list
   this.state			= 0			// Datalink rowset state
   this.modified		= false		// Datalink rowset modified
   this.bookmark		= null		// Datalink rowset bookmark

	*- External Use properties
	this.styleDDL			= true
   this.allowBlank		= false
   this.autoDropOnKey	= false
   this.refreshDDL		= true
	this.refreshKey		= null
	this.picture			= ""
   this.indexName			= null
	this.rangeKey			= null
   this.filter				= null

	*- External Use methods
   this.refresh			= {||False}

   function onOpen
   ***************
      *- Assign refresh function pointer
      if type(this.form.className + "::" + this.name + "_refresh") = "FP"
         cMacro = this.form.className + "::" + this.name + "_refresh"
         this.refresh = &cMacro.
      endif

		*- Get hwnd of Edit control
      this.hwndEdit = FindWindowEx(this.hWnd, 0, null, null)

      *- Set margin of Edit control to 1
      SendMessageN(this.hwndEdit, EM_SETMARGINS, EC_LEFTMARGIN, 1)
      SendMessageN(this.hwndEdit, EM_SETSEL, 0, 0)

		*- Timer object
      this.timer = new timer()
      this.timer.parent    = this
      this.timer.onTimer   = this.onTimer
      this.timer.interval  = 0.05
      this.timer.enabled   = false

      this.timer.dataLink	= false

      return

	function When(bOpen)
   *************
   	this.refresh()

   	if this.refreshDDL AND NOT bOpen
      	this.initDataSource()
      endif

      return true

   function onGotFocus
   *******************
   	if this.autoDropOnKey AND NOT this.autoDrop
      	this.autoDropFlag	= true
      endif

		if TYPE("this.dataLink") == "O"
      	this.timer.dataLink	= true
      endif

		this.modified = false
      this.bookmark = null

      this.timer.enabled = true

      return

   function onLostFocus
   ********************
      this.timer.enabled = false

      return

	function onChange
   *****************
   *- Reset modified property.

      this.modified = false

   	return

   function Key(nChar, nPosition, bShift, bControl)
   ************
   *- Allow, disallow, or modify key input

 		*- AutoDrop list if option set, only once per focus
      // Do not drop on Tab or Delete keys
      if this.autoDropFlag AND (nChar # 8 AND nChar # 9 AND nChar # 46)
         this.DropDown(1)
      endif

		*- Get selected text to determine the resulting string
      local nSelect, nStart, nLength, cInput

      nSelect	= SendMessageN(this.hwndEdit, EM_GETSEL, null, null)
      nStart	= LOWORD(nSelect)
      nLength	= HIWORD(nSelect) - nStart
      
		*- Special key stuffed into buffer to simulate user
      // keypress to ensure control is registered as modified.
      if nChar = 8 AND nStart = 0 AND nLength = 0
      	return true
      endif

		if this.styleDDL

         *- Printable keys, ensure resulting string exists
      	if nChar > 31 AND nChar < 128 AND nChar # 46
		     	cInput	= this.value.stuff(nStart, nLength, CHR(nChar))
            nChar		= this.findString(cInput)

			*- Backspace key, shift selection back left
         elseif nChar = 8
				nChar		= false
				nStart	= MAX(0, nStart - 1)
 
            SendMessageN(this.hwndEdit, EM_SETSEL, nStart, -1)

         *- Delete key, if allowBlank and resulting input string
         // is blank, null value and change nChar to ensure
         // value remains blank (BugFix).
         elseif nChar	= 46
         	nChar		= false
				cInput	= this.value.stuff(nStart, nLength, "")

            if this.allowBlank AND cInput.length = 0
            	nChar = 8
            	this.value = null
            endif

			elseif nChar # 0
         	nChar = false

         endif

		*- Ensure blank values remain blank (as above)
      elseif nChar = 8 OR nChar = 46
      	cInput = this.value.stuff(nStart, nLength, "")

         if cInput.length = 0
				nChar = 8
         	this.value = null
         endif

      endif

      return nChar

	function onKey(nChar, nPosition, bShift, bControl)
   **************
   *- Search for value and perform auto-fill search.

		if nChar = 8 AND nPosition # 1
      	nPosition --
      endif

      if this.FindString(this.value, true)
         SendMessageN(this.hwndEdit, EM_SETSEL, nPosition - 1, -1)
      endif

      this.modified = true

      return

   function onTimer
   ****************
   	this.parent.refresh()

   	if this.dataLink
			this.parent.workaround()
      endif

      return

	function Workaround
   *******************
   *- Workaround two bugs:
   *- The control is not registering as changed when change
   // is made using arrow keys. The datalink value is compared
   // to the control value and when not equal, the control is
   // keyboarded to simulate user input.

   *- Edit control value is not selected when navigating, 
   // saving, or abandoning the datalinked row. The state and
   // bookmark properties are used to evaluate changes in
   // the underlying datalink rowset and an EM_SETSEL message
   // is sent to select the edit value when appropriate.

		local nState, cValue
      nState = this.dataLink.parent.parent.state

      if nState < this.state
         this.modified = false

         SendMessageN(this.hwndEdit, EM_SETSEL, 0, -1)

		elseif this.dataLink.parent.parent.bookmark() # this.bookmark
      	this.bookmark = this.dataLink.parent.parent.bookmark()

         SendMessageN(this.hwndEdit, EM_SETSEL, 0, -1)

      elseif NOT this.modified
      	cValue = this.value.rightTrim()

	   	if NOT this.dataLink.value.rightTrim() == cValue

				// Continue if string exists in dropDownList
         	if this.findString(cValue)
            	SendMessageN(this.hwndEdit, EM_SETSEL, 0, 0)
            	keyBoard(CHR(8))

					// Incase rowset still in browse mode
               if this.dataLink.parent.parent.state = 1
               	this.modified = null

                  SendMessageN(this.hwndEdit, EM_SETSEL, 0, -1)

               endif

            endif
         endif

      endif

      this.state = nState

      return

   function FindString(cString, bSelect)
   *******************
   *- Performs a string search on the dropDown list items,
   // returns true if found.

   *- cString:	String to search
   *- bSelect:	Whether to select the string found

      local nIndex
		nIndex = SendMessageC(this.hWnd, CB_FINDSTRING, -1, cString)

      if bSelect AND nIndex # -1
         SendMessageN(this.hwnd, CB_SETCURSEL, nIndex, 0)
      endif

      return nIndex # -1

   function DropDown(nFlag)
   *****************
   *- Drops down or closes up list window
   *- nFlag: 0 Close Up, 1 Drop, Null Toggle

      local nState
      nState = SendMessageN(this.hwnd, CB_GETDROPPEDSTATE, 0, 0)

      if nFlag = null
         nFlag = IIF(nState = 0, 1, 0)
      endif

      if nFlag # nState
         SendMessageN( this.hwnd, CB_SHOWDROPDOWN, nFlag, 0)
      endif

      this.autoDropFlag = false
      this.curSelFlag	= false

      return

	function InsertString
   *********************
   *- Inserts a string into the dropDown list at specified index.
   *- Params: 	cText - String to Insert
   	//			nIndex - Index number to insert string (0 based)

   	parameters cString, nIndex

      if TYPE("cString") # "C"
      	cString = ""	// Default String
      endif

      if TYPE("nIndex") # "N"
      	nIndex = 0		// Default Index Number (Top)
      endif

      SendMessageC(this.hwnd, CB_INSERTSTRING, nIndex, cString)

		return

	function InitDataSource(bRequery)
   ***********************
   *- Rebuilds the dataSource from a field object, trimming
   // the field values. Uses existing dataSource or dataLink
   // if dataLink has a lookupRowset. The controls indexName,
   // rangeKey, and filter are applied to the field object's
   // rowset if applicable.

   *- Params:
   // bRequery	-	Whether to requery dataSource that uses a
   	//				tempTable. Requery to ensure new rows added
      //				to the original table are included.

      this.refreshDDL = false

      local bInit, cIndex, rSource, fSource, cString
      bInit		= true
      cIndex	= Null

      do case

      case TYPE("this.dataSource.fieldName") # "U"
      	rSource = this.dataSource.parent.parent
         fSource = this.dataSource

         if bRequery
         	this.dataSource.parent.parent.refresh()
         endif

      case TYPE("this.dataLink.lookupRowset.fields[2]") == "O"

			if bRequery
         	if this.dataLink.lookupRowset.tempTable

	         	if EMPTY(this.dataLink.lookupSql)
               	this.dataLink.lookupRowset.parent.requery()

               else
               	this.dataLink.lookupSql = this.dataLink.lookupSql
               endif

            else
              	this.dataLink.lookupRowset.parent.refresh()

            endif

         endif

         rSource = this.dataLink.lookupRowset
         fSource = this.dataLink.lookupRowset.fields[2]

      otherwise
      	bInit = false

      endcase

      if bInit

         *- Empty the current list contents
      	if TYPE("this.dataSource.fieldName") # "U"
         	SendMessageN(this.hwnd, CB_RESETCONTENT, 0, 0)
         else
         	this.dataSource = ""
         endif

         *- Set indexName, range, and filter
         if this.indexName # null

	         cIndex = rSource.indexName
   	      rSource.indexName = this.indexName

         endif

         if this.rangeKey # null
        	   rSource.setRange(this.rangeKey)

         endif

         if this.filter # null
         	rSource.filter = this.filter
         endif

         *- Traverse dataSource rowset to create list items
			if rSource.first()
         	rSource.notifyControls = false

   	      if fSource.type = "CHARACTER"
	
      	   	do
         	   	cString = TRANSFORM(fSource.value.rightTrim(), this.picture)
            		SendMessageC(this.hwnd, CB_ADDSTRING, 0, cString)
            	until NOT rSource.next()

            endif

            rSource.notifyControls = true
         endif

			*- Restore indexName, clear range and filter
         if this.indexName # null
            rSource.indexName = cIndex

         endif

         if this.rangeKey # null
	         rSource.clearRange()
         endif

         if this.filter # null
         	rSource.clearFilter()
         endif

         *- Sync list item with value
         this.findString(this.value, true)

      endif

      return bInit

endclass


