/* 
   FileRecurser.cc
   Ken Chan
   As posted on the dBASE newsgroups, July 7, 2005

   This class will allow you to define a folder, a file filter (if you need one),
   and what to do when you get specific files. It recurses through folders
   starting at the main level, and going through to lower levels. It is a fully
   recursive routine, and (having tested it -- KJM) works quite well.

   This is brilliant, and fast.
   
   March 30, 2016:
   Updated based on code by Ivar B. Jessen in the dBASE newsgroups to deal with
   some versions of language drivers used by dBASE/BDE (specifically the 
   "Borland xxx Latin-1" drivers) and oddities in the way some characters 
   get processed. (Also some work on the indentation and appearance of the 
   code ...)

   Usage:
      set procedure to :dUFLP:FileRecurser.cc additive
      x = new FileRecurser( "C:\WestHistoryWebsite" )
      // only files with .htm as the file extension:
      x.filterFile := {|d,f| lower(right(f,4)) == ".htm"}
      // what to do:
      x.processFile := {|d,f|; AddCSSTag(d + f)}
      // start working ...
      x.search()

   Note that the function shown above in the codeblock for "processFile" 
   calls a function that isn't shown here. If you simply leave this line out
   of the code shown, the default method "processFile" will add each file
   to an array. 

   The filterFile method if overwritten (as shown above) will filter the 
   files processed. If the code above is left out, all files in all folders
   will be processed.
*/

class FileRecurser( cFolderPath )
   //
   // The root directory is C:\, not C:, so always store directories
   // with the trailing (back)slash to be consistent
   //
   local root  // Root of search, not necessarily root of drive...
   root = ltrim(trim( cFolderPath ))
   c = right( root, 1 ) // IBJ
   if asc( c ) # 92     // IBJ 
      root += chr(92) // "\" -- IBJ
   endif
   this.root = root

   protect folders, files
   protect childrenOf
   
   this.folders = null  // List of directories left to search
   this.files   = null  // List of files found so far

   function getFiles
      if this.files == null
         this.search()
      endif
   return this.files

   function search
      this.folders := { this.root }   // Start with the root
      this.files   := new Array()     // To hold files found

      do while this.folders.size > 0
         // Go through everything in the directory and put them
         // in the right list
         this.childrenOf(this.folders[1])
         // Shrink the list of folders to search
         this.folders.delete(1)  // Just did this one
         this.folders.size--
      enddo
   return this.files

   function childrenOf( cFolder )
      local aDir, rows, n, cFile
      aDir = new Array()
      aDir.dir( cFolder + "*.*", "D" )  // Folder names end with "\"

      rows = aDir.size / 5
      for n = 1 to rows
         cFile = aDir[n, 1]
         // Skip entries for current (.) and parent (..) directories
         if cFile == chr(46) or cFile == chr(46)+chr(46) // IBJ
            loop
         endif


         if "D" $ aDir[n, 5]
            // Another folder to search; don't forget trailing "\"
            this.folders.add( cFolder + cFile + chr(92) ) // IBJ
         else
            // File in directory; give the opportunity to filter in/out
            // Could pass other file info here too
            if this.filterFile( cFolder, cFile )
               // Do something with the file
               this.processFile( cFolder, cFile )
            endif // filterFile
         endif // "D" $ ...
      endfor // loop
   return
   
   function filterFile( cFolder, cFile )
      // Override this if desired
   return true

  function processFile( cFolder, cFile )
      // Override this if desired
      // Don't forget to call super::processFile to add to files list,
      // unless you don't want to use the list
      this.files.add( cFolder + cFile )
   return
   
endclass
