/*
	This function will set a new beginning value for the AutoInc column
   in a populated table and alter the existing AutoInc values accordingly.

   Back up your table before using the function!!!

   resetAutoInc( <cTable expC>, <nAuto expN> ) 

	cTable is table name like "Fish.dbf"
	nAuto is an integer setting the beginning value of the new AutoInc values

	Example of usage:
	resetAutoInc( "Fish.dbf" )       // New AutoInc values begin a with 1
   resetAutoInc( "Fish.dbf", 21 )   // New AutoInc values begin a with 21

   *** IMPORTANT NOTE:
       Do not use this on a table where the autoincrement
       field has been used as a foreign key (linking field) --
       the numbers in the autoinc field may be changed, and
       the links will be to the wrong tables (or no longer work).

       This should be used for one of three situations in particular:
       1) an autoinc field that somehow has duplicate values
          (possibly because the field was changed TO an 
          autoincrement field after adding values)
          or
       2) a table with existing data to which an autoincrement
          field was later added
          or
       3) a table where you want to reset the values in the autoincrement
          field, with a new starting point.


       This code will readjust the values in the AutoIncrement
       field, so that they are all sequential. This means that 
       any gaps in numbers will be removed, such as:
           2
           4
           5

       will now be:
           1
           2
           3
       and so on.

       If you pass a starting value, the first record
       will start with that number -- i.e., if you pass 21:
           21
           22
           23 
           ...

       When finished, the 'next number' is adjusted in the
       table header, so that when you add a new record,
       the value will be sequential after the last record's
       value (i.e., n+1).

   This code is not related to "ResetAutoinc.wfm", which only
   adjusts the 'next number' in the table. 
   *** End of Important Note

   Dependencies: None


   Ivar B. Jessen 20-06-2007

*/

function resetAutoInc( cTable, nAuto )

	if empty( argvector( 1 ) )
		return null
	endif

	if empty( argvector( 2 ) )
		nAuto = 1
	endif

	local nRecords, nBytesInHeader, nBytesInRecord
   local nFields, lFlag, cType, nTotalLen, nFieldNo
   local NewNextAutoValue, n1

	f = new file()
	f.open( cTable, "RW" )
	f.seek( 4, 0 )

// Read number of records in table, bytes 4-7
	nRecords = Read32()

// Read number of bytes in header, bytes 8-9
	nBytesInHeader = Read16()

// Read number of bytes in record, bytes 10-11
	nBytesInRecord = Read16()

/* The number of fields is: 
	The number of bytes in the header, 
   minus the number of bytes from the beginning of the header 
   to the beginning of the Field Descriptor Array at byte 68,
   minus 1 byte because the byte count starts with byte 0,
   all divided by 48, as each row of the Field Descriptor Array is 48 bytes long
*/

	nFields = int( ( nBytesInHeader - 1 - 68 )/48 ) 

/*
	Now let us find the autoinc field which is of type "+"
	The field type in the Field Descriptor Array comes after the Name of the 
	field, which is 32 bytes long. But we also need to know the lengths of the
	fields preceeding the AutoInc field. Those field lengths are found one byte
   downstream of the field types in the Field Descriptor Array.
*/

	f.seek( 68 + 32, 0 )
				
   lFlag = false
	nTotalLen = 1
	for i = 1 to nFields
		cType = ( asc( f.read( 1 ) ) )	// Field type
	   if cType = 43 							// chr(43) = "+"
	   	nFieldNo = i
         lFlag = true
	   	exit
   	endif
	   nTotalLen += asc( f.read( 1 ) ) // Field length
	   f.seek( 46, 1 ) // Next row in field array is 48 bytes further on minus one
   						 // byte for the nextAutoInc number and one byte for the 
                  	 // field length
	next i

	if lFlag = false  // No autoInc field in table
   	return null
	endif

 // Go to the byte containing the Next AutoInc value.
	f.seek( 68 + 1 + ( nFieldNo - 1 ) * 48 + 41, 0 )

// Calculate the new value for the next AutoInc value, which is
// start value ( nAuto ) plus number of records in table 
	NewNextAutoValue = ( nAuto + nRecords )

// Write it into the table
	f.write( writeLongB( NewNextAutovalue ) )

// Goto first AutoInc field below End of File Header,
// which field begins at nBytesInHeader + nTotalLen
	f.seek( nBytesInHeader + nTotalLen, 0 )

// Insert replacement AutoInc values
	n1 = nAuto
	for i = 1 to nRecords
  		f.write( writeLong( n1 ) )
		n1++
		f.seek( nBytesInRecord - 4, 1 )
	next i

	f.close()

return 

//------------- Functions -------------

function Read16
	return ASC(f.read(1))+ASC(f.read(1))*256

function Read32
	return ASC(f.read(1))+ASC(f.read(1))*256+ASC(f.read(1))*256*256+;
   ASC(f.read(1))*256*256*256

function ReadLong				// Read 32-bit signed numeric
   local n
   n = ReadULong()
   return n-2147483648

function ReadULong			// Read 32-bit unsigned numeric
   local c
   c = f.Read(4)
   return Int(Asc(SubStr(c, 4)) +;
   BitLShift(Asc(SubStr(c, 3)),  8) +;
   BitLShift(Asc(SubStr(c, 2)), 16) +;
   BitLShift(Asc(SubStr(c, 1)), 24))


function WriteLong(n)
	// Write 32-bit signed numeric
   return WriteULong(IIf(n < 0, n+2147483648, n-2147483648))


function WriteULong(n)
	// Write 32-bit unsigned numeric
   nV = (;
   Chr(BitAnd(BitRShift(n, 24), 0xFF)) +;
   Chr(BitAnd(BitRShift(n, 16), 0xFF)) +;
   Chr(BitAnd(BitRShift(n,  8), 0xFF)) +;
   Chr(BitAnd(n, 0xFF)))
   return nV


FUNCTION WriteLongB(n)
	// Write 32-bit signed numeric
   RETURN WriteULongB(IIf(n > 0, n, 0x100000000 + n))


FUNCTION WriteULongB(n)
	// Write 32-bit unsigned numeric
   RETURN (Chr(BitAnd(n, 0xFF)) +;
			 Chr(BitAnd(BitRShift(n,  8), 0xFF)) +;
			 Chr(BitAnd(BitRShift(n, 16), 0xFF)) +;
			 Chr(BitAnd(BitRShift(n, 24), 0xFF)))



