#define FileVersion 3.0
/*
    ADOSeqValue.cc
    Programmer: Ken Mayer
    Date......: May, 2015
    Notes.....: This is based on the concepts in the old XDML
                SEQValue.cc class that Romain (Strieff) created 
                ages past, updated specifically for SQL Server
					 databases, for The dBASE Book Plus.

                This is a copy of ADOSqlValue.cc created for
					 the book, changing the data objects to standard
					 OODML (Database and Query, rather than ADODatabase
					 and ADOQuery). Everything else should work the same.
					 
                This class requires the Write Database Connnection,
					 name of the table, the name of the field, and 
					 then it will attempt to perform the autoincrement.
					 It uses a table (SEQValue), if it does not exist,
					 the table will be created. 
					 
					 You can maintain as many autincrement values
					 as needed with this class, but only one per table.

    Usage:      set procedure to SQLSeqValue.cc
                oSeq = new SeqValue( <database alias>, <loginString> )
					 // For MSSQL databases, you may be required to use 
					 // "dbo.TableName here ...
                oSeq.Increment( <tablename>, <fieldname> )
                oSeq.Reset( <tablename>, <fieldname>, <newValue> )
                oSeq.release()

                Example:
                set proc to SQLSeqValue.cc
                oSeq = new SeqValue( "MyDatabase", "Username/password" )
                nValue = oSeq.Increment( "Invoices", "InvoiceNumber" )
                oSeq.release()

    Methods:    InitTable() // Create table if needed, otherwise
                            // open table used to store/increment
                            // values as a property of the
                            // class  ** PROTECTED **
                release()   // as it sounds, should be called to
                            // release the object and close things down.
                Increment() // performs the increment, requires
                            // you pass the tablename and the fieldname
                            // as parameters.
                            // Example:
                            //    oSeq.Increment( "MyTable", "MyField" )
                Reset()     // Resets the autoinc value in a field.
                            // Important: to start at 1, you must
                            // reset to zero. 
                            // Example:
                            //    oSeq.Reset( "MyTable", "MyField", 0 )

    Side Effects: This creates the SeqValue table in the database.
    
*/
class SeqValue
   protect InitTable
   if pCount() < 2
      msgbox( "Need database name and loginString as a parameter", ;
		        "SQLSeqValue.cc Error", 16 )
		cancel
   else
      this.DataBase    = ArgVector( 1 ) // we have a database alias
		this.loginString = ArgVector( 2 ) // we have a loginString
   endif

   // startup code:
   if not class::InitTable()
      cancel // stop processing
   endif

   function InitTable
      local oIndex, cDataBase, cLoginString
      private cCmd

      // Create the table if it does not (yet) exist:
      cDatabase = this.database
		cLoginString = this.loginString
      this.lDBCreated = false
		try
			this.oDB               = new Database()
         this.oDB.databaseName := cDatabase
			this.oDB.loginString  := cLoginString
			this.oDB.active       := true
			this.lDBCreated       := true
		catch( ADOException E )
			msgbox( "Error: "+e.code+" - "+e.message+chr(13)+;
					  "Line: "+e.LineNo,;
					  "SQLSeqValue -- Error setting up database", 16 )
			return false
		endtry

      // Check to see if the table exists:
      if not this.oDB.tableExists( "SeqValue" )
         // create table
         cCmd = [create table "SEQVALUE" (]+;
                [IncTable  char( 32 ),]+;
                [IncField  char( 32 ),]+;
                [IncValue  bigint )]
         this.oDB.executeSQL( cCmd )
      endif

      // Create query for table:
      this.SeqTable           = new Query()
      this.SeqTable.database := this.oDB
      this.SeqTable.sql      := "select * from SeqValue"
      this.SeqTable.active   := true
      // place us at the first row:
      this.SeqTable.rowset.first()

   return true
   // End of Method: InitTable()

   function release
      // release the query object after setting
      // the active property to false
      this.SeqTable.active := false
      release object this.SeqTable
      // if the database object was created:
      if this.lDBCreated
         this.oDB.active := false
         release object this.oDB
      endif
      // and release this object, and we're done.
      release object this
   return null

   function Increment( cTable, cField )
      // here's where we actually deal with incrementing
      // the value ...

      // error checking for required parameters
      if pCount() < 1
         msgbox( "You must pass a tablename to be used"+;
                 "for the increment process.",;
                 "SQLSeqValue Error", 16 )
         return
      endif
      if pCount() < 2
         msgbox( "You must pass a fieldname to be used"+;
                 "for the increment process.",;
                 "SQLSeqValue Error", 16 )
         return
      endif

      // moving on, we need to see if the table and field
      // names are in the table:
      local bAddRow
      bAddRow = false
      this.SeqTable.rowset.filterOptions := 2 // ignore case
      this.SeqTable.rowset.filter        := [IncTable = ']+cTable+[']
      if not this.SeqTable.rowset.endOfSet
         // we're not at the end of the rowset, so
         // look for the fieldname:
         this.SeqTable.rowset.locateOptions := 2 // ignore case
         if this.SeqTable.rowset.applyLocate( [IncField = ']+cField+['] )
            // we found it!
            this.SeqTable.rowset.fields["IncValue"].value++
            this.SeqTable.rowset.save()
            return this.SeqTable.rowset.fields["IncValue"].value
         else
            // the field doesn't exist, but the table is
            // referenced, so we need to still add a row:
            bAddRow := true
         endif
      else
         // The table is not referenced, so we need to add a row:
         bAddRow := true
      endif

      // if we need to add the row:
      if bAddRow
         this.SeqTable.rowset.beginAppend()
         this.SeqTable.rowset.fields["IncTable"].value := cTable
         this.SeqTable.rowset.fields["IncField"].value := cField
         this.SeqTable.rowset.fields["IncValue"].value := 1
         this.SeqTable.rowset.save()
         return 1 // we created a new row, starting at 1
      endif
      // there shouldn't need to be an "else", as at this point,
      // we should have trapped everything, and be set ...
   return "Error" // just for completeness' sake
   // End of Method: Increment()

   function Reset( cTable, cField, nNewVal )
      // Reset the value to 'nNewVal'

      // error checking for required parameters
      if pCount() < 1
         msgbox( "You must pass a tablename to be used "+;
                 "for the increment process.",;
                 "SQLSeqValue Error", 16 )
         return "Error"
      endif
      if pCount() < 2
         msgbox( "You must pass a fieldname to be reset ...",;
                 "SQLSeqValue Error", 16 )
         return "Error"
      endif
      if pCount() < 3
         msgbox( "You must pass a new value to be used "+;
                 "for the increment field.",;
                 "SQLSeqValue Error", 16 )
         return "Error"
      endif

      private n
      n = nNewVal
      if type( "n" ) <> "N"
         msgbox( "Value to be used must be numeric "+;
                 "(for the increment field).",;
                 "SQLSeqValue Error", 16 )
         return "Error"
      endif

      if nNewVal < 0
         msgbox( "Value to be used must be 0 or greater "+;
                 "(for the increment field).",;
                 "SQLSeqValue Error", 16 )
         return "Error"
      endif
      
      // moving on, we need to see if the table and field
      // names are in the table:
      local bRowFound
      bRowFound = false
      this.SeqTable.rowset.filterOptions := 2 // Ignore case
      this.SeqTable.rowset.filter := [IncTable = ']+cTable+[']
      if not this.SeqTable.rowset.endOfSet
         // we're not at the end of the rowset, so
         // look for the fieldname:
         this.SeqTable.rowset.locateOptions := 2 // Ignore case
         if this.SeqTable.rowset.applyLocate( [IncField = ']+cField+['] )
            // we found it!
            bRowFound := true
            this.SeqTable.rowset.fields["IncValue"].value := nNewVal
            this.SeqTable.rowset.save()
            return this.SeqTable.rowset.fields["IncValue"].value
         else
            // We didn't find it ... 
            bRowFound := false
         endif
      else
         // We didn't find it ...
         bRowFound := false
      endif

      // if we need to add the row:
      if not bRowFound
         msgbox( "Did not find table '"+cTable+"' with field '"+cField+"'."+chr(13)+;
                 "Cannot reset field value.",;
                 "SQLSeqValue Error", 16 )
         return "Error!"
      endif
      // there shouldn't need to be an "else", as at this point,
      // we should have trapped everything, and be set ...
   return "Error" // just for completeness' sake
   // End of Method: Increment()

endclass

