// Metric: 0 - Chars or 6 - Pixels only
// TIMEFLD.CC

/*
   TimeEntryfield 
   Author: Peter Rorlick
   Programming contributions by Serge Plourde

   DEPENDENCIES: TIME_CFG.WFM

   // minor modifications for dUFLP library -- renamed
   // class to "TimeEntryField" rather than "TimeField",
   // and formatted the comments a bit. Code is untouched.
   // KJM, 7/5/98

   This is a very smart entryfield that morphs a HHMM C,4 field 
   into a nicely readable 12-hour or 24-hour format.  It also 
   lets the user type in a time in almost any format, interprets 
   the user's shorthand, and redisplays the input in the 
   normalized format. For example, type "230p" and it will display
   "2:30 p.m." (or "14:30" if the app is in 24-hour mode) and it 
   will store "1430" in the DBF field.


  When datlinked, field morphing is done automatically so that 
  the user sees the time in a nice format rather than just HHMM.
  
  There are 2 display modes:
    24-hour military (e.g., "23:59")
    12-hour am/pm  (e.g., "11:59 p.m.")
  The display mode that is used will depend on an _app.TimeFormat 
  variable (default is 12-hour if _app.TimeFormat doesn't exist).
  
  The user is able to type in the time in a variety of formats, and 
  the field "interprets" (morphs) what the user has typed, and 
  redisplays it in the appropriate format. If the user has entered 
  something that is truly nonsensical, an error message box is issued.
  The idea is to be very flexible in letting the user enter the time 
  in whatever format he/she desires. Here are some examples:
  
     What the user can type:  Interpretion in military time:
     ----------------------   -----------------------------
     ?                        Current time clock value
     now                      Current time clock value
     23:59			            23:59
     2359			   			23:59
     234			      			02:34
     234p			   			14:34
     11:30pm		   			23:30
     11:30p		      			23:30
     11:30			   			11:30
     11pm			   			23:00
     11			      			11:00
     931p			   			21:31
     8765xyz		   			error
 
  To show a C,4 time field in a grid, you should create a calculated 
  field whose beforeGetValue is 
        {||DisplayTime(this.parent['YourTimeField'].value)}
  and show that calculated field in the grid rather than the real 
  field.
 
  Related functions from Funclib.prg:
    NormalTimeStrToMilitary(cNormalizedTimeString)
    NormalizeTimeString(cTimeString)
    MilitaryDisplay(cMilitaryTime)
    MilitaryTo12hour(cMilitaryTime)
    DisplayTime(cMilitaryTime)
   --------------------------------------------------------------------
*/

class TimeEntryfield(f) of EntryField(f) custom

   with (this)
      onRightMouseDown  := {; do :dUFLP:time_cfg.wfm with .t.; this.keyboard(this.value+chr(13)) }
      speedtip          := 'Right-click to change format'
      picture           := 'XXXXXXXXXX'
      width             := 10
      validErrorMessage := 'Invalid time'
   endwith

   function onDesignOpen( lFromPalette )
      if lFromPalette and form.metric == 6 // Pixels
         this.height := 22
         this.width  := 70
      endif
   return

   Proc OnOpen
      if empty(this.datalink)
         // No datalink:
	      this.OnChange = class::timeEntryField_OnChange
      else
         // The entryfield is Datalinked:
         this.datalink.canChange = TimeEntryField::CanChange
         this.datalink.beforeGetValue = TimeEntryField::beforeGetValue
         form.rowset.refreshControls()
      endif


   Func CanChange(NewValue)
      // Assumption: newValue is a character string of up to 
      // 10 characters.
      local cTimeString, cHours, cMinutes
      cTimeString = NormalizeTimeString(newValue)
      if cTimeString == 'INVALID'
         beep()
         msgbox('Invalid time! ' + newValue, 'Error', 16)
      else
         this.value = NormalTimeStrToMilitary(cTimeString)
      endif
   return false


   Proc TimeEntryField_OnChange
      local cTimeString, cHours, cMinutes
      cTimeString = NormalizeTimeString(this.Value)
      if cTimeString == 'INVALID'
         beep()
         msgbox('Invalid time! ' + this.Value, 'Error', 16)
      else
         this.value = DisplayTime( NormalTimeStrToMilitary(cTimeString) )
      endif


   Func BeforeGetValue
      // Assumptions:
      //	 -The time is stored in 4 char in the field in military 
      //       format HHMM.
      //  -The time in the dbf field is expected to be valid.
      if this.parent.parent.endOfSet
         return null                     // at end of rowset
      elseif this.value == null
        // For beginAppend()
        return ""
      else
         if empty(this.value)
            return ''
         endif
      endif
      return NormalizeTimeString(this.value)

endclass

/*
  --------------------------------------------------------------
  Beginning of related functions                               
  --------------------------------------------------------------
*/

Func DisplayTime(cHHMM)
/*
   Author: Peter Rorlick

   cHHMM must be 4 or 5 characters, in military format. The ":" is 
   optional. Returns a string in one of two formats, e.g. "11:59 p.m." 
   or "23:59 p.m." The latter (military format) is used only when 
   _app.TimeFormat is a string but not "12-hour".

   This function is ideal for displaying a time field in a grid.
   Example:
    function times1_onOpen
      local f
      f = new field()
      f.fieldname = 'Start time'
      f.beforeGetValue = {||DisplayTime(this.parent['StartTime'].value)}
      this.rowset.fields.add(f)
      return

  (For entryfields datalinked to time fields, use the 
  TimeField class in Time.cc)
  
*/
   if empty(cHHMM)
      return space(10)
   elseif type('_app.TimeFormat') # 'C' or ;
          Upper(_app.TimeFormat) == '12-HOUR'
      // 12-hour display:
      return padr(MilitaryTo12hour( left(cHHMM,2) + ':' +;
             right(cHHMM,2) ), 10)  
   else
      // 24-hour display:
      return padr(MilitaryDisplay( left(cHHMM,2) + ':' + right(cHHMM,2) ), 10) 
   endif


FUNC MilitaryDisplay(cMilitaryTime)   && Returns time string in military format + am/pm.
   /*
      Assumptions:
         -cMilitaryTime MUST BE A VALID 24-hour HH:MM format 
   */
   LOCAL nHH, cFirstChar

   IF EMPTY(cMilitaryTime)
      RETU ''
   ENDIF

   IF LEN(cMilitaryTime) <> 5 or AT(':', cMilitaryTime) <> 3
      RETU 'INVALID'
   ENDIF

   nHH = VAL(LEFT(cMilitaryTime,2))

   IF LEFT(cMilitaryTime,1) = '0'
      cMilitaryTime = SUBS(cMilitaryTime,2)
   ENDIF

RETU cMilitaryTime + IIF(nHH>11,'',' a.m.')

FUNC MilitaryTo12hour(cMilitaryTime)   && Returns time string in 12-hour format
   /*
      Assumptions:
         -cMilitaryTime MUST BE either HH:MM or HHMM or empty
         -cMilitaryTime is in VALID 24-hour format
   */
   IF EMPTY(cMilitaryTime)
      RETU ''
   ENDIF
   IF LEN(cMilitaryTime) <> 5 or AT(':', cMilitaryTime) <> 3
     RETU 'INVALID'
   ENDIF

   LOCAL nHH, cRetVal
   nHH = VAL(LEFT(cMilitaryTime,2))

   IF AT(':', cMilitaryTime) = 0
     cMilitaryTime = LEFT(cMilitaryTime,2) + ':' + RIGHT(cMilitaryTime,2)
   ENDIF

   IF nHH > 12 &&add pm indicator here rather than at end
     cRetVal = LTRIM(STR(nHH-12)) + RIGHT(cMilitaryTime,3) + ' p.m.'
   ELSEIF nHH = 12 &&additional check for noon
     cRetVal = cMilitaryTime + ' p.m.'
   ELSEIF nHH = 0 &&additional check for midnight
     cRetVal = '12' + RIGHT(cMilitaryTime,3) + ' a.m.'
   ELSE
     cRetVal = LTRIM(STR(nHH)) + RIGHT(cMilitaryTime,3) + ' a.m.'
   ENDIF
   RETU cRetVal

Func NormalizeTimeString(cTimeString)
   /*
      Author: Serge Plourde
      Created on 98.04.20

      'Normalize' a given time string into a format such as 11:00 p.m. 
      or 23:00.

      cTimeString may look like:
       ?  or now     Current time clock value
       23:59		   23:59
       2359			   23:59
       234			   2:34
       234p			   14:34
       11:30pm		   23:30
       11:30p		   23:30
       11:30		   11:30
       11pm			   23:00
       11			   11:00
       931p			   21:31
       8765xyz		   error: 8765 bad time value

   */

   local cNormalizedTimeString
   local lResumeNormalization
   local nAt  //Position of ':' in time string
   local cHours
   local cMinutes
   local cCharacter
   local cAmOrPm
   private cHoursAndMinutes

   if empty(cTimeString)
      return ''
   endif

   lResumeNormalization = true

   //First remove any blank from the time string:
   cNormalizedTimeString = allTrim(cTimeString)
   cNormalizedTimeString = StrTran(cNormalizedTimeString, ' ', '' )
   cNormalizedTimeString = StrTran(cNormalizedTimeString, '.', '' )
   if cNormalizedTimeString = '?' or upper(left(cNormalizedTimeString,3)) == 'NOW'
      //Replace '?' with current system time clock value:
      if type('_app.TimeFormat')#'C' or Upper(_app.TimeFormat)=='12-HOUR'
         return militaryTo12Hour(left(time(),5))
      else
         return militaryDisplay(left(time(),5))
      endif
   else
      if not (left(cNormalizedTimeString,1)$'1234567890')
         return 'INVALID'  //invalid: it must start with a digit.
      endif
   endif

   //Parse the time string for the hours and minutes
   nAt = At(':', cNormalizedTimeString)
   cHours = ''
   if nAt <> 0
      //Parse hours and minutes from both sides of ':'
      if nAt = 1
         cHours = '00'
      else
         cHours = left(cNormalizedTimeString, nAt - 1)
         if len(cHours) = 1 
            cHours = '0' + cHours
         endif
      endif

      //Parse the minutes
      if nAt < len(cNormalizedTimeString)
         cNormalizedTimeString = subs(cNormalizedTimeString, nAt + 1)
      else
         cNormalizedTimeString = ''
      endif

      cMinutes = ''      
      do while true and not empty(cNormalizedTimeString)
         //will exit as soon as no minutes value from string or empty string
         cCharacter = left(cNormalizedTimeString,1)
         if cCharacter $ '1234567890'
            cMinutes = cMinutes + cCharacter
            if len(cNormalizedTimeString) > 1
               cNormalizedTimeString = subs(cNormalizedTimeString,2)
            else
               cNormalizedTimeString = ''
               exit
            endif
         else
            exit
         endif
      enddo

      if empty(cMinutes)
         cMinutes = '00'
      else
         if len(cMinutes) = 1
            cMinutes = '0' + cMinutes
         endif
      endif
    
   endif //':' in time string: parsed hours and minutes

   if empty(cHours)
      //means that ':' was not part of the string
      cHoursAndMinutes = ""
      do while len(cHoursAndMinutes) < 5 and not empty(cNormalizedTimeString)
         //will exit as soon as no minutes value from string
         cCharacter = left(cNormalizedTimeString,1)
         if cCharacter $ '1234567890'
            cHoursAndMinutes = cHoursAndMinutes + cCharacter
            if len(cNormalizedTimeString) > 1
               cNormalizedTimeString = subs(cNormalizedTimeString,2)
            else
               cNormalizedTimeString = ''
               exit
            endif
         else
            exit
         endif
      enddo


      do case
      case len(cHoursAndMinutes) = 1
         cHours = '0' + cHoursAndMinutes
         cMinutes = '00'

      case len(cHoursAndMinutes) = 2
         cHours = cHoursAndMinutes
         cMinutes = '00'
      case len(cHoursAndMinutes) = 3
         cHours = '0' + left(cHoursAndMinutes, 1)
         cMinutes = right(cHoursAndMinutes, 2)
      case len(cHoursAndMinutes) = 4
         cHours = left(cHoursAndMinutes, 2)
         cMinutes = right(cHoursAndMinutes, 2)
      otherwise
         return 'INVALID'
      endcase
   endif

   /*
   At this point we have cHours and cMinutes in HHMM with leading zero.
   cNormalizedTimeString contains the remaining of the given string after the hours
   and minutes have been parsed.
   */
   cAmOrPm = upper(left(ltrim(cNormalizedTimeString),1))
   if cHours = '12' and cAmOrPm <> 'A'
      cAmOrPm = 'P'
   endif
   if empty(cAmOrPm) 
      cAmOrPm = 'A'
   endif

   if cAmOrPm = 'P'   //time labeled as pm
      if val(cHours) < 12
         cHours = str(val(cHours) + 12, 2)
      endif
   *->Test source code:
   else
      *-AM
      if cHours = '12'
         cHours = '00'
      endif
   *<-End of test source code
   endif

   if Val(cHours) > 24 or val(cMinutes) > 60 or;
      (cHours + cMinutes >= '2400')
      return 'INVALID'  //invalid time
   endif

   if type('_app.TimeFormat')#'C' or Upper(_app.TimeFormat)=='12-HOUR'
      //return 12-hour format
      cNormalizedTimeString = militaryTo12Hour( cHours + ':' + cMinutes)
   else
      //24-hour format
      cNormalizedTimeString = militaryDisplay(cHours + ':' + cMinutes)
   endif
 
return cNormalizedTimeString


Func NormalTimeStrToMilitary(cTimeString)
   /*
   Author: Serge Plourde
   Created on 98.04.22

   Take a 'Normalize(d)TimeString() and returns a military time.
   */
   local cHours, cMinutes

   if empty(cTimeString)
      return spac(4)
   endif

   if type('_app.TimeFormat')#'C' or Upper(_app.TimeFormat)=='12-HOUR'
      /*
      cTimeString is in 12-hour mode: 12:00 a.m. @ 12:59 a.m. are 
      actually 00:00 @ 00:59
      */
      do case
      case '12:' $ cTimeString and 'a.m.' $ cTimeString
         cHours = '00'
      case 'p.m.' $ cTimeString and not '12:' $ cTimeString
         cHours = left(cTimeString,at(':', cTimeString) - 1)           
         cHours = str(val(cHours) + 12, 2)
      case 'p.m.' $ cTimeString and left(cTimeString,3) == '12:'
         cHours = '12'
      otherwise
         cHours = left(cTimeString,at(':', cTimeString) - 1)           
      endcase
   else
      *-cTimeString is already in 24-hour mode
      cHours = left(cTimeString,at(':', cTimeString) - 1)
   endif
   cMinutes = subs(cTimeString, at(':', cTimeString) + 1, 2)
    	  
   if len(cHours) = 1
      cHours = '0' + cHours
   endif
return cHours + cMinutes

/*
   --------------------------------------------------------------------
   End of functions related to Time fields                     
   --------------------------------------------------------------------
*/   

func alltrim(cstring)
return ltrim(rtrim(cstring))

* Pad-right function, for clipper/Foxpro compatibility.
Func padr(cString,nLength,cPadChar)
return left( cString+REPL(IIF(PCOUNT()<3,' ',cPadChar),nLength), nLength )

&& Like Clipper's built-in function.
Function StrTran( cString, cSearch, cRepl)     

   LOCAL nAt, cString1, Handle  // Hanlde - Numeric value that indicate number of character in string that already parsed...
   Priv cReplace

   local cReturnString

   cReturnString=cString

   cReplace=cRepl

   IF TYPE('cReplace') <> 'C'  && If this optional parameter was omitted...
     cReplace = ""  && Omiting the 3rd parameter means "remove all instances of the searh string".
   ENDIF

   Handle=0
   cString1=cReturnString
   nAt = AT(cSearch,cString1)
   DO WHILE nAt > 0
     cString1 = STUFF(cString1, nAt, LEN(cSearch), cReplace)
     cReturnString =  SubStr(cReturnString, 1,Handle) + cString1
     Handle = Handle + nAt + Len(cReplace)-1
     cString1 = SubStr(cReturnString, Handle+1,Len(cReturnString) - Handle)
     nAt = AT(cSearch,cString1)
   ENDDO
RETURN cReturnString

Func Beep
   param nFreq,nDur,cWAVFilename
   * Params are optional. First 2 are ignored if cWAVFilename is passed.
   if type('cWAVFilename')=='C'   // if this param was passed...
     play sound filename &cWAVFilename   // instead of chr(7)
     return
   endif
   private cSaveSettings, cNewSettings
   cSaveSettings = setTo('BELL')
   if type('nFreq')=='N'   // if this param was passed...
     cNewSettings = ltrim(str(nFreq)) + ',' + ltrim(str(nDur))
   else
     cNewSettings = '800,20'   // Default settings (applied when no params are passed).
   endif
   set bell to &cNewSettings
   ?? chr(7)
   set bell to &cSaveSettings
return

/*
   --------------------------------------------------------------------
   End of file: TimeFld.cc
   --------------------------------------------------------------------
*/
